@@ -15,28 +15,29 @@
 (*  Based on Mldvi by Alexandre Miquel.                                *)
 (***********************************************************************)
 
-(* $Id: drawimage.ml,v 1.7 2003/01/17 12:57:16 furuse Exp $ *)
+(* $Id: drawimage.ml,v 1.18 2004/10/01 11:13:35 weis Exp $ *)
 
 let debugs = Misc.debug_endline;;
 
-type ratiopts =
+type white_is_transparent = bool;;
+
+type ratiopt =
    | ScaleOriginal    (* leave the image at its original native size  *)
    | ScaleAuto        (* scale to fit requested area                  *)
    | ScaleCenter      (* scale as needed to cover the image,
-                         keep original Ratio and center *)
+                         keep original ratio and center *)
    | ScaleTop         (* scale x coords to align to top of the screen,
-                         keep original Ratio  *)
+                         keep original ratio  *)
    | ScaleBottom      (* scale x coords to align to bottom of the screen,
-                         keep original Ratio  *)
+                         keep original ratio  *)
    | ScaleLeft        (* scale y coords to align to left of the screen,
-                         keep original Ratio  *)
+                         keep original ratio  *)
    | ScaleRight       (* scale y coords to align to right of the screen,
-                         keep original Ratio  *)
+                         keep original ratio  *)
    | ScaleTopLeft
    | ScaleBottomLeft
    | ScaleTopRight
-   | ScaleBottomRight
-;;
+   | ScaleBottomRight;;
 
 (* Blending *)
 type blend =
@@ -44,7 +45,27 @@ type blend =
    | ColorDodge | ColorBurn | Darken | Lighten | Difference
    | Exclusion (* | Luminosity | Color | Saturation | Hue *);;
 
-type alpha = float;;
+let string_of_blend = function
+  | Normal -> "Normal"
+  | Multiply -> "Multiply"
+  | Screen -> "Screen"
+  | Overlay ->"Overlay"
+  (*
+  | SoftLight -> "SoftLight"
+  | HardLight -> "HardLight"
+  *)
+  | ColorDodge -> "ColorDodge"
+  | ColorBurn -> "ColorBurn"
+  | Darken -> "Darken"
+  | Lighten -> "Lighten"
+  | Difference -> "Difference"
+  | Exclusion -> "Exclusion"
+  (*
+  | Luminosity -> "Luminosity"
+  | Color -> "Color"
+  | Saturation -> "Saturation"
+  | Hue -> "Hue"
+  *);;
 
 (* look at gxblend.c of ghostscript *)
 let blend_func = function
@@ -104,33 +125,41 @@ let blend_func = function
         let t = t + t lsr 8 in
         t lsr 8;;
 
+type alpha = float;;
+
 type image_size = int * int;;
 
-type ps_bbox = int * int * int * int;; 
+type position = int * int;;
 
-open Image
-open Color
-open GraphicsY11
+type antialias = bool;;
+
+type ps_bbox = int * int * int * int;;
+
+open Images;;
+open Color;;
+open GraphicsY11;;
 
 let cache_prefix = "cache";;
 let cache_key = "advicache";;
 
-let verbose_image_access = Options.flag false
+let verbose_image_access =
+  Options.flag false
     "--verbose-image-access"
-    "\tChange the cursor while image loadings";;
+    "  change the cursor while loading images,\
+    \n\t (the default is to keep the same cursor).";;
 
-let image_aa = 
+let image_aa =
   Options.flag true
     "-disable-image-anti-aliasing"
-    "\tDisable eps inclusion anti-aliasing"
-;;
+    "  disable eps inclusion anti-aliasing,\
+    \n\t (the default is to use anti-aliasing).";;
 
 let images_camlimages = Hashtbl.create 107;;
 let images_graphics = Hashtbl.create 107;;
 
-let after f g = try let x = f () in  g (); x with e -> g (); raise e;;
+let after f g = try let x = f () in  g (); x with exc -> g (); raise exc;;
 
-let cache_path file whitetransp psbbox ratiopt (w, h) =
+let cache_path file whitetransp psbbox ratiopt antialias (w, h) =
   let file' = Userfile.fullpath (Unix.getcwd ()) file in
   let file' = if file == file' then String.copy file' else file' in
   for i = 0 to String.length file' - 1 do
@@ -171,7 +200,7 @@ let cache_path file whitetransp psbbox ratiopt (w, h) =
        bbox_str
        ratiopt_str
        (if whitetransp then "-tr" else "")
-       (if !image_aa then "-aa" else ""))
+       (if antialias then "-aa" else ""))
 ;;
 
 let cache_load file =
@@ -179,8 +208,7 @@ let cache_load file =
   let load ic =
     let s = String.create (String.length cache_key) in
     ignore (input ic s 0 (String.length cache_key));
-    if s <> cache_key then
-      raise (Failure (file ^ " has no proper header"));
+    if s <> cache_key then failwith (file ^ " has no proper header");
     let rgba = input_value ic in (* bool *)
     let width = input_value ic in
     let height = input_value ic in
@@ -192,8 +220,9 @@ let cache_load file =
       open_in_bin, close_in
     else
       (fun file ->
-        let command = Printf.sprintf "%s -c -d %s" Config.gzip_path 
-	    (Filename.quote file) in
+        let command =
+          Printf.sprintf "%s -c -d %s" Config.gzip_path (Filename.quote file) in
+        debugs ("cache_load " ^ command);
         Unix.open_process_in command),
       (fun ic -> ignore (Unix.close_process_in ic))
   in
@@ -226,12 +255,11 @@ let cache_save file img =
     | _ -> assert false
   in
   let opener, closer =
-    if Config.gzip_path = "" then
-      open_out_bin, close_out
-    else
+    if Config.gzip_path = "" then open_out_bin, close_out else
       (fun file ->
-        let command = Printf.sprintf "%s > %s" Config.gzip_path 
-	    (Filename.quote file) in
+        let command =
+          Printf.sprintf "%s > %s" Config.gzip_path (Filename.quote file) in
+        debugs ("cache_save " ^ command);
         Unix.open_process_out command),
       (fun oc -> Pervasives.flush oc; ignore (Unix.close_process_out oc))
   in
@@ -258,46 +286,36 @@ let image_load psbbox (w, h) file =
   debugs ("image load " ^ file);
   let image_aa_level = if !image_aa then 2.0 else 1.0 in
   try
-    match fst (Image.file_format file) with
-    | Image.Ps ->
+    match fst (Images.file_format file) with
+    | Images.Ps ->
         let (llx, lly, urx, ury) =
         match psbbox with
         | Some bbox -> bbox
-        | None -> (0, 0, w, h)
-        in
+        | None -> (0, 0, w, h) in
         (* we need anti-aliasing *)
         let resx = float w /. (float (urx - llx) /. 72.0) *. image_aa_level
         and resy = float h /. (float (ury - lly) /. 72.0) *. image_aa_level
         in
-
-        (* resolution fix *)
-        (* gs dies if resolutions are too small (around 1.6 - 1.65dpi) *)
+        (* resolution fix:
+           gs dies if resolutions are too small (around 1.6 - 1.65dpi) *)
         let gs_min_res = 2.0 in
-        let resx = 
-          if resx < gs_min_res then gs_min_res
-          else resx
-        and  resy = 
-          if resy < gs_min_res then gs_min_res
-          else resy
-        in
-        
+        let resx = max resx gs_min_res
+        and  resy = max resy gs_min_res in
         Ps.load_ps file (Some (llx, lly, urx, ury))
           [Load_Resolution (resx, resy)]
-    | _ -> Image.load file []
+    | _ -> Images.load file []
   with
-  | Failure s ->
-      Misc.warning ("Failed to load " ^ file ^ ": " ^ s);
-      raise Exit
-  | Image.Wrong_file_type ->
+  | Images.Wrong_file_type ->
       Misc.warning ("Unsupported graphic format in: " ^ file);
       raise Exit
-  | _ ->
-      Misc.warning ("Failed to load " ^ file);
+  | exc ->
+      Misc.warning
+        (Printf.sprintf "Failed to load %s: %s" file (Printexc.to_string exc));
       raise Exit
 ;;
 
 let resize_and_make_transparent image whitetransp ratiopt (ow, oh) =
-  let iw, ih = Image.size image in
+  let iw, ih = Images.size image in
   let w, h =
     (* compute the size of the scaled image according to the options *)
     match ratiopt with
@@ -348,19 +366,20 @@ let resize_and_make_transparent image whitetransp ratiopt (ow, oh) =
     | ScaleBottomRight ->
         Some(max 0 (w - ow), max 0 (h - oh), 0, 0, min w ow, min h oh)
   in
-  let image' =
+  (* Coerce all formats to Rgba 24 or Rgba32 *)
+  let image =
     match image with
     | Index8 i -> Rgba32 (Index8.to_rgba32 i)
     | Index16 i -> Rgba32 (Index16.to_rgba32 i)
     | Rgb24 i -> image
     | Rgba32 i -> image
-    | _ -> raise (Failure "color model is not supported")
-  in
+    | _ -> failwith "color model is not supported" in
+  (* Adding transparency if necessary *)
   let white_rgb = {r = 255; g = 255; b = 255} in
   let diff = 15 * 15 in
-  let image'' =
-    if not whitetransp then image' else
-    match image' with
+  let image =
+    if not whitetransp then image else
+    match image with
     | Rgb24 i ->
         let width = i.Rgb24.width
         and height = i.Rgb24.height in
@@ -369,8 +388,8 @@ let resize_and_make_transparent image whitetransp ratiopt (ow, oh) =
           for x = 0 to width - 1 do
             let rgb = Rgb24.unsafe_get i x y in
             let a =
-              if whitetransp && Rgb.square_distance rgb white_rgb < diff 
-	      then 0 else 255 in
+              if whitetransp && Rgb.square_distance rgb white_rgb < diff
+              then 0 else 255 in
             Rgba32.unsafe_set rgba x y { color = rgb; alpha = a }
           done
         done;
@@ -388,7 +407,8 @@ let resize_and_make_transparent image whitetransp ratiopt (ow, oh) =
         image
     | _ -> assert false
   in
-  match image'' with
+  (* Resizing the image then bliting it at the proper place. *)
+  match image with
   | Rgb24 i ->
       let i = Rgb24.resize None i w h in
       let i =
@@ -405,72 +425,77 @@ let resize_and_make_transparent image whitetransp ratiopt (ow, oh) =
         match blitinfo with
         | None -> i
         | Some(sx, sy, dx, dy, w, h) ->
-            let i' = Rgba32.make ow oh {color = white_rgb; alpha = 1} in
+            let i' = Rgba32.make ow oh {color = white_rgb; alpha = 255} in
             Rgba32.blit i sx sy i' dx dy w h;
             i' in
       Rgba32 i
   | _ -> assert false
 ;;
 
-let load_and_resize file whitetransp psbbox ratiopt (w, h) =
+let load_and_resize file whitetransp psbbox ratiopt antialias (w, h) =
   let file =
     Search.true_file_name [ "-format=" ^ Config.texpicts_kind ] file in
-  let cache_name = cache_path file whitetransp psbbox ratiopt (w, h) in
+  let cache_name =
+    cache_path file whitetransp psbbox ratiopt antialias (w, h) in
   let image =
     try
-      try
-        Hashtbl.find images_camlimages cache_name
-      with
+      try Hashtbl.find images_camlimages cache_name with
       | Not_found ->
           if (Unix.stat file).Unix.st_mtime >=
-             (Unix.stat cache_name).Unix.st_mtime
-          then raise Exit;
+             (Unix.stat cache_name).Unix.st_mtime then raise Exit;
           (* Use the cache file *)
-          if !verbose_image_access
-            then GraphicsY11.set_cursor Cursor_exchange;
+          if !verbose_image_access then GraphicsY11.set_cursor Cursor_exchange;
           let im = cache_load cache_name in
-          if !verbose_image_access
-            then GraphicsY11.set_cursor Cursor_left_ptr;
-          im
-    with _ ->
-      if !verbose_image_access then GraphicsY11.set_cursor Cursor_exchange;
+          if !verbose_image_access then GraphicsY11.set_cursor Cursor_left_ptr;
+          im with
+    | _ ->
+        if !verbose_image_access then GraphicsY11.set_cursor Cursor_exchange;
         begin try Unix.unlink cache_name with _ -> () end;
         let image = image_load psbbox (w, h) file in
         let image =
           resize_and_make_transparent image whitetransp ratiopt (w, h) in
         Userfile.prepare_file cache_name;
-        (* we have no trivial image format for rgba32! *)
         cache_save cache_name image;
         if !verbose_image_access then GraphicsY11.set_cursor Cursor_left_ptr;
-        image
-  in
+        image in
   Hashtbl.replace images_camlimages cache_name image;
   cache_name, image
 ;;
 
 let draw_image image cache_name alpha blend (w, h) (x0, y0) =
+  debugs (Printf.sprintf "Alpha is %d" alpha);
   let blend =
-    try Some (blend_func blend) with _ -> None in
+    try Some (blend_func blend) with _ -> debugs "No blend"; None in
   (* load_and_resize may not return exactly the same size
      we specified as (w, h) *)
-  let iw, ih = Image.size image in
+  let iw, ih = Images.size image in
   match image with
-  | Rgb24 _ when alpha = 1.0 && blend = None ->
+  | Rgb24 _ when alpha = 255 && blend = None ->
       (* optimized *)
       let image_graphics =
-        try Hashtbl.find images_graphics cache_name with
+        try
+          let img = Hashtbl.find images_graphics cache_name in
+          debugs ("cached graphics image found for " ^ cache_name);
+          img with
         | Not_found ->
+            debugs ("no cached graphics image for " ^ cache_name);
             let im = Graphic_image.of_image image in
             Hashtbl.add images_graphics cache_name im;
             im
       in
       Graphics.draw_image image_graphics x0 y0
   | Rgb24 _ | Rgba32 _ ->
-      let alpha = truncate (alpha *. 255.0) in
       let get_src_alpha =
         match image with
-        | Rgb24 image -> fun x y -> Rgb24.unsafe_get image x y, alpha
+        | Rgb24 image ->
+            debugs (
+             Printf.sprintf "Rgba 24 with blend or alpha (%d) %s"
+              alpha cache_name);
+            fun x y -> Rgb24.unsafe_get image x y, alpha
         | Rgba32 image ->
+            debugs (
+             Printf.sprintf "Rgba 32 with blend or alpha (%d) %s"
+              alpha cache_name);
             fun x y ->
               let {color = src; alpha = a} = Rgba32.unsafe_get image x y in
               src, a * alpha / 255
@@ -508,15 +533,17 @@ let draw_image image cache_name alpha blend (w, h) (x0, y0) =
   | _ -> assert false
 ;;
 
-
-let f file whitetransp alpha blend psbbox ratiopt (w, h) (x0, y0) =
+let f file whitetransp alpha blend psbbox ratiopt antialias (w, h) (x0, y0) =
   try
+    let antialias = if not !image_aa then false else antialias in
     let cache_name, image =
-      load_and_resize file whitetransp psbbox ratiopt (w, h) in
+      load_and_resize file whitetransp psbbox ratiopt antialias (w, h) in
     (* load_and_resize may not return exactly the same size
        we specified as (w, h) *)
-    debugs (Printf.sprintf "Draw %s (alpha=%f blend=%s)" file
-              alpha "Don't know");
+    debugs
+      (Printf.sprintf "Draw %s (alpha=%f blend=%s)"
+         file alpha (string_of_blend blend));
+    let alpha = truncate (alpha *. 255.0) in
     draw_image image cache_name alpha blend (w, h) (x0, y0);
     debugs "Success Draw!";
   with
@@ -524,5 +551,4 @@ let f file whitetransp alpha blend psbbox ratiopt (w, h) (x0, y0) =
   | Unix.Unix_error (e, s1, s2) ->
       Misc.warning (Printf.sprintf "%s %s: %s" s1 s2 (Unix.error_message e))
   | Failure s -> Misc.warning s
-  | e -> Misc.warning (Printexc.to_string e)
-;;
+  | e -> Misc.warning (Printexc.to_string e);;
