@@ -15,10 +15,11 @@
 (*  Based on Mldvi by Alexandre Miquel.                                *)
 (***********************************************************************)
 
-(* $Id: dviview.mli,v 1.16 2003/01/06 10:49:11 weis Exp $ *)
+(* $Id: dviview.mli,v 1.18 2003/10/28 12:43:28 remy Exp $ *)
 
 open Dimension;;
 val set_autoresize : bool -> unit;;
+val set_autoscale : bool -> unit;;
 val set_geometry : string -> unit;;
 val set_crop : bool -> unit;;
 val set_hmargin : dimen -> unit;;
@@ -26,4 +27,4 @@ val set_vmargin : dimen -> unit;;
 
 exception Error of string;;
 
-val main_loop : string -> unit;;
+val main_loop : string -> string list -> unit;;
