@@ -15,6 +15,11 @@
 (*  Based on Mldvi by Alexandre Miquel.                                *)
 (***********************************************************************)
 
-(* $Id: shot.mli,v 1.2 2003/01/06 10:49:14 weis Exp $ *)
+(* $Id: shot.mli,v 1.5 2004/03/04 17:27:44 weis Exp $ *)
 
-val save : unit -> unit;;
+type x = int and y = int and w = int and h = int;;
+
+val save_page_area_image_file : Misc.file_name -> x -> y -> w -> h -> unit;;
+val save_page_area_image : x -> y -> w -> h -> unit;;
+val save_page_image_file : Misc.file_name -> unit;;
+val save_page_image : unit -> unit;;
